<?php

namespace Drupal\iack\Form;

use Drupal\Core\Link;
use Drupal\Core\Url;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleExtensionList;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\File\FileUrlGeneratorInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Markup;
use Drupal\file\Entity\File;
use Drupal\iack\Service\IackHelper;
use PhpOffice\PhpSpreadsheet\IOFactory;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Create a form to upload excel sheet with information architecture.
 */
class UploadInformationArchitectureForm extends FormBase {

  /**
   * {@inheritdoc}
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager.
   * @param \Drupal\Core\File\FileSystemInterface $fileSystem
   *   File system service.
   * @param \Drupal\iack\Service\IackHelper $iackHelper
   *   Iack helper service.
   * @param \Drupal\Core\Extension\ModuleExtensionList $moduleExtensionList
   *   The module extension list.
   * @param \Drupal\Core\File\FileUrlGeneratorInterface $fileUrlGenerator
   *   The file URL generator.
   */
  public function __construct(
    protected EntityTypeManagerInterface $entityTypeManager,
    protected FileSystemInterface $fileSystem,
    protected IackHelper $iackHelper,
    protected ModuleExtensionList $moduleExtensionList,
    protected FileUrlGeneratorInterface $fileUrlGenerator,
  ) {
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('file_system'),
      $container->get('iack.iack_helper'),
      $container->get('extension.list.module'),
      $container->get('file_url_generator')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId(): string {
    return 'iack_upload_information_architecture_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {

    $prompts = $form_state->get('prompts');
    if (!empty($prompts)) {
      $form['ai_assistant'] = [
        '#type' => 'hidden',
        '#value' => $form_state->get('ai_assistant'),
      ];
      foreach ($prompts as $key => $prompt) {
        $form[$key] = [
          '#type' => 'textarea',
          '#title' => $key,
          '#default_value' => $prompt,
          '#required' => TRUE,
          '#rows' => '8',
          '#description' => $this->t('Edit prompt before sending it to AI agents.'),
        ];
      }

      $form['actions']['process'] = [
        '#type' => 'submit',
        '#value' => $this->t('Human Confirmation'),
        '#submit' => [
          [$this, 'processForm'],
        ],
      ];
      // A container for displaying the message.
      $form['message'] = [
        '#type' => 'markup',
        '#markup' => '<div id="ajax-form-message"></div>',
      ];
    }
    else {
      $disabled = FALSE;
      if (!class_exists('PhpOffice\PhpSpreadsheet\IOFactory')) {
        $this->messenger()->addError($this->t('Required PhpOffice library not found.'));
        $disabled = TRUE;
      }

      $ai_assistants = $this->entityTypeManager->getStorage('ai_assistant')->loadMultiple();
      $ai_assistants_options = [];
      foreach ($ai_assistants as $key => $ai_assistant) {
        $ai_assistants_options[$key] = $ai_assistant->label();
      }

      if (!empty($ai_assistants_options)) {
        $form['ai_assistant'] = [
          '#type' => 'select',
          '#title' => $this->t('AI Assistants'),
          '#description' => $this->t('Please select an AI assistant to create Information Architecture.'),
          '#options' => $ai_assistants_options,
        ];
      }
      else {
        $this->messenger()->addError($this->t('Please configure at least one AI assistant.'));
        $disabled = TRUE;
      }

      $path = $this->moduleExtensionList->getPath('iack') . '/template/iatemplate.xlsx';
      $url = $this->fileUrlGenerator->generate($path);
      $url = Url::fromUri('base://' . $path, [
        'attributes' => [
          'target' => '_blank',
        ],
      ]);

      // Build a link object from the URL.
      $link = Link::fromTextAndUrl($this->t('Download template'), $url)->toString();
      $download_template = '<p>Use this template to create Information architecture:</p>' . $link;
      $form['download_template'] = [
        '#markup' => Markup::create($download_template),
      ];
      $form['information_architecture_file'] = [
        '#type' => 'managed_file',
        '#title' => $this->t('Upload an information architecture file'),
        '#description' => $this->t('File types allowed: xls xlsx'),
        '#upload_validators' => [
          'FileExtension' => [
            'extensions' => 'xls xlsx',
          ],
        ],
        '#upload_location' => 'temporary://iack/',
        '#required' => TRUE,
      ];

      $form['actions']['submit'] = [
        '#type' => 'submit',
        '#value' => $this->t('Submit'),
      ];

      if ($disabled) {
        $form['actions']['submit']['#attributes'] = [
          'disabled' => 'disabled',
        ];
      }
    }
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {

    // Get the file data.
    $file_id = $form_state->getValue('information_architecture_file')[0];
    if (empty($file_id)) {
      return;
    }
    $ai_assistant = $form_state->getValue('ai_assistant');
    $file = File::load($file_id);
    $file_path = $file->getFileUri();

    // Ensure the file is accessible.
    $real_path = $this->fileSystem->realpath($file_path);
    $sheets_data = [];

    try {
      // Load the Excel file.
      $spreadsheet = IOFactory::load($real_path);

      // Get all sheets and their data.
      foreach ($spreadsheet->getAllSheets() as $sheet) {
        $sheet_name = $sheet->getTitle();
        $sheet_data = $sheet->toArray();
        // Ignore headers.
        unset($sheet_data[0]);
        // Filter out empty rows.
        $filtered_data = array_filter($sheet_data, function ($row) {
          // Remove rows with empty strings and whitespace.
          return array_filter($row, function ($cell) {
            return $cell !== NULL && trim($cell) !== '';
          });
        });

        // Prepare the sheets data keyed by sheet name.
        $sheets_data[$sheet_name] = $filtered_data;
      }
    }
    catch (\Exception $e) {
      $this->messenger()->addError($this->t('An error occurred while parsing the Excel file: @message', ['@message' => $e->getMessage()]));
    }

    // Parse sheet data to get AI prompt.
    $prompts = $this->iackHelper->getAiPrompt($sheets_data);
    $form_state->set('prompts', $prompts);
    $form_state->set('ai_assistant', $ai_assistant);
    $form_state->setRebuild();
  }

  /**
   * {@inheritdoc}
   */
  public function processForm(array &$form, FormStateInterface $form_state) {
    $ai_assistant = $form_state->getValue('ai_assistant');

    $prompts = array_filter($form_state->getValues(), function ($value, $key) {
      if (strpos($key, 'prompt') !== FALSE) {
        return $value;
      }
    }, ARRAY_FILTER_USE_BOTH);

    // Define the batch operations.
    $operations = [];
    // Process prompts in batches.
    foreach ($prompts as $prompt) {
      $operations[] = [
        [$this, 'processPrompt'],
        [$prompt, $ai_assistant],
      ];
    }

    // Define the batch.
    $batch = [
      'title' => $this->t('Creating your Information Architecture...'),
      'operations' => $operations,
      'finished' => [$this, 'batchFinish'],
      'progress_message' => $this->t('Processed @current out of @total.'),
      'error_message' => $this->t('An error occurred during processing.'),
    ];

    // Set the batch.
    batch_set($batch);
  }

  /**
   * Batch API callback for calling ai assistant api.
   *
   * @param string $prompt
   *   AI prompt.
   * @param string $ai_assistant
   *   AI assistant id.
   * @param mixed|array $context
   *   The batch current context.
   */
  public function processPrompt($prompt, $ai_assistant, &$context) {
    // Send prompt to ai assistant.
    try {
      $ai_response = $this->iackHelper->processPrompt($prompt, $ai_assistant);
      if (!isset($context['results'])) {
        $context['results'] = [];
      }
      $context['results'] = array_merge($context['results'], [$ai_response]);
    }
    catch (\Exception $e) {
      $this->messenger()->addError($this->t('An error occurred while processing prompt: @message', ['@message' => $e->getMessage()]));
    }
  }

  /**
   * Batch API finished.
   *
   * @param bool $success
   *   TRUE if batch successfully completed.
   * @param array $results
   *   Batch results.
   * @param array $operations
   *   An array of function calls (not used in this function).
   */
  public function batchFinish($success = FALSE, array $results = [], array $operations = []) {
    $response = implode("\n", $results);
    if (!$success) {
      $this->messenger()->addError($this->t('Finished with an error.'));
    }
    else {
      $this->messenger()->addStatus($this->t('AI Response:') . $response);
    }
  }

}
