<?php

namespace Drupal\iack\Service;

use Drupal\ai_assistant_api\AiAssistantApiRunner;
use Drupal\ai_assistant_api\Data\UserMessage;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\ai_assistant_api\Entity\AiAssistant;

/**
 * The iack helper.
 */
class IackHelper {

  /**
   * Constructs a new IackHelper object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager.
   * @param \Drupal\ai_assistant_api\AiAssistantApiRunner $aiAssistantClient
   *   The AI Assistant API client.
   */
  public function __construct(
    protected EntityTypeManagerInterface $entityTypeManager,
    protected AiAssistantApiRunner $aiAssistantClient,
  ) {
  }

  /**
   * Parse the excel sheet in into AI prompt.
   *
   * @param array $sheets_data
   *   The agent.
   */
  public function getAiPrompt(array $sheets_data) {
    $prompts = [];

    // Create Vocabulary prompt.
    if (!empty($sheets_data["Taxonomy terms"])) {
      $group_terms = [];
      foreach ($sheets_data["Taxonomy terms"] as $row) {
        $group_key = trim($row[0]);
        $group_terms[$group_key][] = $row;
      }
      foreach ($group_terms as $vocabulary => $terms) {
        // Create Vocabulary.
        $prompt = '';
        if (FALSE !== $key = array_search($vocabulary, array_column($sheets_data["Vocabulary"], 0))) {
          $prompt .= 'Create Taxonomy vocabulary - ' . $vocabulary;
          // @todo Support machine name and description.
        }
        $prompt .= "\nAdd the following terms to the $vocabulary vocabulary:\n";
        $terms = array_column($terms, 1);
        $prompt .= implode(', ', $terms);
        $prompts[$vocabulary . ' Vocabulary prompt'] = $prompt;
      }
    }

    // Create content type prompt.
    if (!empty($sheets_data["Fields"])) {
      $group_fields = [];
      foreach ($sheets_data["Fields"] as $row) {
        $group_key = trim($row[0]);
        $group_fields[$group_key][] = $row;
      }
      foreach ($group_fields as $bundle => $fields) {
        // Create content type.
        $prompt = '';
        if (FALSE !== $key = array_search($bundle, array_column($sheets_data["Bundles"], 0))) {
          $prompt .= 'Create content type - ' . $bundle;
          if (!empty($sheets_data["Bundles"][$key + 1][1])) {
            $prompt .= ", Machine name: " . $sheets_data["Bundles"][$key + 1][1];
          }
          if (!empty($sheets_data["Bundles"][$key + 1][2])) {
            $prompt .= ", Description: " . $sheets_data["Bundles"][$key + 1][2];
          }
        }

        $prompt .= "\nAdd the following fields to the $bundle content type:\n";
        foreach ($fields as $field) {
          $prompt .= "- Label: " . $field[1];
          if (!empty($field[2])) {
            $prompt .= ", Machine name: " . $field[2];
          }
          if (!empty($field[3])) {
            $prompt .= ", Field type: " . $field[3];
          }
          if (!empty($field[4])) {
            $prompt .= ", Field widget: " . $field[4];
          }
          if (!empty($field[5])) {
            $prompt .= ", Reference Entity: " . $field[5];
          }
          $prompt .= "\n";
        }
        $prompts[$bundle . ' Content type prompt'] = $prompt;
      }
    }

    return $prompts;
  }

  /**
   * Send the Prompt to AI agent to process.
   */
  public function processPrompt(string $prompt, string $ai_assistant) {
    // Load the Ai Assistant entity.
    $assistant = $this->entityTypeManager->getStorage('ai_assistant')->load($ai_assistant);
    if (!$assistant instanceof AiAssistant) {
      throw new \UnexpectedValueException('AI Assistant not found.');
    }

    // Set the assistant in the AiAssistantApiRunner.
    $this->aiAssistantClient->setAssistant($assistant);

    $this->aiAssistantClient->streamedOutput(FALSE);
    $user_message = new UserMessage($prompt);
    $this->aiAssistantClient->setUserMessage($user_message);
    $this->aiAssistantClient->setThrowException(TRUE);

    // Process the user's message.
    try {
      $response = $this->aiAssistantClient->process();

      // Handle the response.
      $normalizedResponse = $response->getNormalized();
      return $this->createResponse($normalizedResponse);
    }
    catch (\Exception $e) {
      return $e->getMessage();
    }
  }

  /**
   * Returns a normal response.
   */
  protected function createResponse($normalizedResponse) {
    $assistantResponseText = $normalizedResponse->getText();
    // Set the assistant message for logging or further processing.
    $this->aiAssistantClient->setAssistantMessage($assistantResponseText);

    $converter = $this->getCommonMarkConverter();
    $assistantResponseText = $this->rewriteMarkdownMessage($assistantResponseText);
    $assistantResponseText = $converter ? $converter->convert($assistantResponseText) : $assistantResponseText;

    return $assistantResponseText;
  }

  /**
   * Rewrite markdown message.
   *
   * @param string $message
   *   The message to rewrite.
   *
   * @return string
   *   The rewritten message.
   */
  protected function rewriteMarkdownMessage(string $message): string {
    // Find any link from a markdown message an iterate them.
    $pattern = '/\[(.*?)\]\((.*?)\)/';
    preg_match_all($pattern, $message, $matches, PREG_SET_ORDER);
    foreach ($matches as $match) {
      $link = trim($match[2]);
      $text = $match[1];
      // If the link does not start with a protocol or slash, add the base URL.
      if (strpos($link, '://') === FALSE && strpos($link, '/') !== 0) {
        $link = base_path() . $link;
      }
      // Sometimes it adds double slashes at the start.
      if (substr($link, 0, 2) === '//') {
        $link = substr($link, 1);
      }

      $message = str_replace($match[0], "[$text]($link)", $message);
    }
    return $message;
  }

  /**
   * Gets the common mark converter if available.
   *
   * @return \League\CommonMark\CommonMarkConverter|null
   *   The common mark converter.
   */
  protected function getCommonMarkConverter() {
    $converter = NULL;
    if (class_exists('League\CommonMark\CommonMarkConverter')) {
      // Ignore the non-use statement loading since this dependency may not
      // exist.
      // @codingStandardsIgnoreLine
      $converter = new \League\CommonMark\CommonMarkConverter();
    }
    return $converter;
  }

}
